from parser import Parser
from common import State, Character, Eps, Terminal, Nonterminal
from dfa import Composition as comp
from typing import Set, Dict, Tuple, Union
from copy import deepcopy
from antlr4 import RecognitionException
from reg import RegGrammar, DFA, NFA, RegEx


def make_dfa(states: Set[str], characters: Set[str],
             transition: Dict[Tuple[str, str], str], init: str, final: Set[str]):

    new_init = State(init)
    new_states: Set[State] = set()
    new_states.add(new_init)
    dfa = DFA(new_states, set(), {}, new_init, set())

    for name in states:
        state = State(name)
        dfa.states.add(state)
    for name in characters:
        character = Character(name)
        dfa.characters.add(character)
    for name in final:
        state = State(name)
        dfa.final.add(state)
    for x, y in transition:
        state = State(x)
        character = Character(y)
        move = State(transition[x, y])
        dfa.transition[state, character] = move

    dfa.check()
    return dfa


def reg_1():
    S = Nonterminal("S")
    A = Nonterminal("A")
    a = Terminal("a")
    nonterminals = {S, A}
    init = S
    terminals = {a}
    rules : Dict[Nonterminal, Set[Union[Terminal, Tuple[Terminal, Nonterminal]]]] = dict()
    rules[S] = {(a, S), (a)}
    rules[A] = {(a)}
    reg = RegGrammar(nonterminals, terminals, rules, init)
    return reg


def reg_2():
    S = Nonterminal("S")
    A = Nonterminal("A")
    a = Terminal("a")
    nonterminals = {S, A}
    init = S
    terminals = {a}
    rules : Dict[Nonterminal, Set[Union[Terminal, Tuple[Terminal, Nonterminal]]]] = dict()
    reg = RegGrammar(nonterminals, terminals, rules, init)
    return reg

def nfa_1():
    q0 = State("q0")
    q1 = State("q1")
    q2 = State("q2")
    a = Character("a")
    transition = {(q0, a): {q1, q2}, (q1, a): {q1}, (q2, a): {q2}}
    nfa = NFA({q0, q1, q2}, {a}, transition, q0, {q1, q2})
    return nfa

def nfa_e():
    q0 = State("q0")
    q1 = State("q1")
    q2 = State("q2")
    q3 = State("q3")
    q4 = State("q4")
    a = Character("a")
    b = Character("b")
    c = Character("c")
    d = Character("d")
    e = Eps()
    transition: Dict[Tuple[State, Union[Character, Eps]], Set[State]] = \
                {(q0, e): {q1}, (q1, a): {q0}, (q1, d): {q3}, (q1, e): {q2},
                 (q2, a): {q3}, (q3, e): {q4}, (q4, b): {q3}, (q4, c): {q2}}
    nfae = NFA({q0, q1, q2, q3, q4}, {a, b, c, d}, transition, q0, {q2})
    return nfae

def dfa_eq(str1, str2):
    try:
        parser = Parser()
        dfa1 = parser.str_to_dfa(str1)
        print(parser.dfa_to_str(dfa1))
        dfa2 = parser.str_to_dfa(str2)
        print(parser.dfa_to_str(dfa2))
        return DFA.is_equivalent(dfa1, dfa2)

    except RecognitionException as re:
        print(re)

def regex_test(str):
    parser = Parser()
    ast = parser.str_to_regex(str)
    str1 = parser.regex_to_str(ast)
    #ast1 = parser.str_to_reg(str1)
    print(str)
    print(str1)
    #print(parser.reg_to_str(ast1))
    efa = ast.regex_to_efa()
    print(parser.nfa_to_str(efa))
    print()

def grammar_test(str):
    parser = Parser()
    gra = parser.str_to_cfg(str)
    str1 = parser.cfg_to_str(gra, True)
    #ast1 = parser.str_to_reg(str1)
    print(str)
    print(str1)
    print()

def main():
    dfa_1 = make_dfa({"q_0", "q_1"}, {"a"}, {("q_0", "a"): "q_1", ("q_1", "a"): "q_1"}, "q_0", {"q_1"})
    dfa_2 = make_dfa({"r_0", "r_1", "r_2"}, {"a", "b"}, {("r_0", "a"): "r_1", ("r_0", "b"): "r_1",
                                       ("r_1", "a"): "r_2", ("r_1", "b"): "r_2"}, "r_0", {"r_2"})
    dfa_3 = make_dfa({"1", "2", "3", "4", "5", "6", "7"}, {"a", "b"}, {("1", "a"): "2", ("2", "a"): "3", ("2", "b"): "4",
                    ("3", "a"): "6", ("3", "b"): "5", ("4", "a"): "3", ("4", "b"): "2", ("5", "a"): "6", ("5", "b"): "3",
                    ("6", "a"): "2", ("7", "a"): "6", ("7", "b"): "1", }, "1", {"3", "5", "6"})

    reg1 = reg_1()
    reg2 = reg_2()
    nfa1 = nfa_1()
    parser = Parser()

    print("parser.dfa_to_str():")
    print("String representation of DFA, simple/full.")
    print()
    print(parser.dfa_to_str(dfa_1))
    print()
    print(parser.dfa_to_str(dfa_1, True))
    print()

    print("______________")
    print("DFA.accepts():")
    print("Test of DFA accepting given word:")
    print("dfa_1.accepts('a'):", dfa_1.accepts("a"))
    print("dfa_1.accepts('ab'):", dfa_1.accepts("ab"))
    print()

    print("________________________________")
    print("DFA.make_total(), DFA.is_total()")
    print("Add new terminal for DFA, check its totality, then make total.")
    print()
    b = Terminal("b")
    dfa_1.characters.add(b)
    print("dfa_1.terminals.add(b), dfa_1.is_total():", dfa_1.is_total())
    print()
    total_dfa = dfa_1.total()
    print("dfa_1.total(), total_dfa.is_total():", total_dfa.is_total())
    print()
    print(parser.dfa_to_str(total_dfa, True))
    print()

    print("________________")
    print("DFA.complement()")
    print("Complement of DFA.")

    compl_dfa = dfa_1.complement()
    print(parser.dfa_to_str(compl_dfa, True))
    print()

    print("_________________")
    print("DFA.composition()")
    print("Composition makes union, intersection and subtraction of two DFAs.")

    print()
    print("second DFA:")
    print(parser.dfa_to_str(dfa_2, True))
    print()

    print("Composition:")
    union = dfa_1.composition(dfa_1, dfa_2, comp.Union)
    print(parser.dfa_to_str(union, True))
    print()
    print("Intersection:")
    intersection = dfa_1.composition(dfa_1, dfa_2, comp.Intersection)
    print(parser.dfa_to_str(intersection, True))
    print()

    print("___________________________")
    print("DFA.eliminate_unreachable()")
    print("Add new state (also to transition) and eliminate unreachable states.")

    dfa_x = deepcopy(dfa_1)
    q_x = State("q_x")
    a = Terminal("a")
    dfa_x.states.add(q_x)
    dfa_x.final.add(q_x)
    dfa_x.transition[q_x, a] = q_x
    print()
    print("DFA with new state:")
    print(parser.dfa_to_str(dfa_x, True))
    print()
    print("DFA without unreachable states:")
    eliminated = dfa_x.eliminate_unreachable()
    print(parser.dfa_to_str(eliminated, True))
    print()

    print("______________")
    print("DFA.minimize()")
    print("Minimization of DFA from slides:")
    print()
    print(parser.dfa_to_str(dfa_3, True))

    print()
    print("Minimal DFA:")
    minimal = dfa_3.minimize()
    print(parser.dfa_to_str(minimal, True))
    print()

    print("______________")
    print("DFA.canonize()")
    print("Canonization of previous minimized DFA.")
    print()
    canonical = minimal.canonize()
    print(parser.dfa_to_str(canonical, True))
    print()

    print("__________________________________")
    print("DFA.dfa_to_reg(), REG.reg_to_nfa()")
    print("Transformation of DFA to REG and REG to NFA")
    print("plus string representation of REG and NFA.")
    print()
    print("Original DFA:")
    dfa_4 = make_dfa({"q_0", "q_1"}, {"a"}, {("q_0", "a"): "q_1", ("q_1", "a"): "q_1"}, "q_0", {"q_1"})
    print(parser.dfa_to_str(dfa_4, True))
    print()
    dfareg = dfa_4.dfa_to_reggrammar()
    print("DFA transformed to REG:")
    print(parser.reggrammar_to_str(dfareg, True))
    print()
    regdfa = dfareg.reggrammar_to_nfa()
    print("REG transformed to NFA:")
    print(parser.nfa_to_str(regdfa, True))
    print()

    print("__________________________________")
    print("DFA.is_empty(), DFA.is_equivalent()")
    print("Emptiness with counterexample and equivalence \n"
          "of DFA through intersections with complements.")
    print()
    print("is dfa3 empty?", dfa_3.is_empty())
    print("dfa3 and canonical dfa3: ", DFA.is_equivalent(dfa_3, canonical), ", should be True")
    print()

    print("__________________________________")
    print("NFA.determinize()")
    print("Determinization of NFA to DFA:")
    det = nfa1.determinize()
    print(parser.nfa_to_str(nfa1, True))
    print()
    print(parser.dfa_to_str(det, True))

    print("__________________________________")
    print("NFA.eliminate_epsilon()")
    print("Elimination of epsilon steps in NFA.")
    nfae = nfa_e()
    print(parser.nfa_to_str(nfae))
    print()
    nfa_elim = nfae.eliminate_epsilon()
    print(parser.nfa_to_str(nfa_elim))

    print("__________________________________")
    print("Testing parsers")
    print("Input of parser.str_to_dfa:")
    print("init=0 (0,x)=1 (0,y)=2 (1,x)=1 (1,y)=1 (2,x)=0 (2,y)=4 (3,x)=2 (3,y)=5 (4,x)=3 (4,y)=0 (5,x)=5 (5,y)=5 final={4,5}")
    dfa_p = parser.str_to_dfa("init=0 (0,x)=1 (0,y)=2 (1,x)=1 (1,y)=1 (2,x)=0 (2,y)=4 (3,x)=2 (3,y)=5 (4,x)=3 (4,y)=0 (5,x)=5 (5,y)=5 final={4,5}")
    print("Output of parser.dfa_to_str should be identical:")
    print(parser.dfa_to_str(dfa_p))
    print()
    print("Try equality of automata from strings:")
    str1 = "init=0 (0, a)=1 (0, b)=2 (1, a)=1 (2,b)=2 final={1,2}"
    str2 = "init=q0 (q0, a)=q1 (q0, b)=q2 (q1,a)=q1 (q2,b)=q3 (q3,b)=q3 final={q1, q2, q3}"
    print(dfa_eq(str1, str2).left_counterexample, dfa_eq(str1, str2).right_counterexample)
    print()

    print("Input of parser.str_to_nfa:")
    print("init=0 (0,x)={1} (0,y)={2} (1,x)={1} (1,y)={1} (2,x)={0,1,2} (2,y)={1,2} final={2,3}")
    nfa_p = parser.str_to_nfa("init=0 (0,x)={1} (0,y)={2} (1,x)={1} (1,y)={1} (2,x)={0,1,2} (2,y)={1,2} final={2,3}")
    print("Output of parser.nfa_to_str should be identical:")
    print(parser.nfa_to_str(nfa_p))
    print()
    print("Parsing EFA:")
    print("init=0 (0,x)={1} (0,\e)={2} (1,x)={1} (1,y)={1} (2,\e)={0,1,2} (2,b)={1,2} final={2,3}")
    efa_p = parser.str_to_nfa("init=0 (0,x)={1} (0,\e)={2} (1,x)={1} (1,y)={1} (2,\e)={0,1,2} (2,b)={1,2} final={2,3}")
    print("Output of parser.nfa_to_str should be identical:")
    print(parser.nfa_to_str(efa_p))
    print()

    print("__________________________________")
    print("Testing regex")
    regex_test("(c.(b.b*.(c+ab)+cb)*.(ε+a))+b")
    regex_test("b")
    regex_test("c+b")
    regex_test("ε")
    regex_test("c.(b.b*c*a*.(b*+ac)+b).a")
    regex_test("b*a*c*")
    regex_test("b^+a^+c^+")
    regex_test("b^+ + a^+ + c^+")
    regex_test("bac")
    regex_test("(b)(a)(c)")
    regex_test("b+a+c+a+c")
    regex_test("∅")
    regex_test("\e")
    regex_test("(∅ + ∅^*).(∅^* + ∅∅^+)")

    print("__________________________________")
    print("More transitions warning")
    ts = parser.str_to_nfa("init=0 (0,a)={0} (0,a)={1} (1,a)={1} final={1}")
    print(parser.nfa_to_str(ts))
    print()

    print("__________________________________")
    print("Parsing grammars")
    grammar_test("S->a|aA\nA->a|b")
    grammar_test("S->a|aA|\e\nA->a|b|<aBcJK50l>\n<aBcJK50l>->AaBa")

main()