from __future__ import annotations
from typing import Set, FrozenSet, List, Dict, Union, Tuple, Deque, Optional, TypeVar, Any
from enum import Enum
import enum
from copy import deepcopy
from collections import deque
from string import ascii_uppercase
from common import Character, State, Terminal, Nonterminal


class Composition(Enum):
    Union = enum.auto()
    Intersection = enum.auto()
    Subtraction = enum.auto()


class IsEmptyResult:
  def __init__(self, counterexample : Optional[str] = None):
    self.counterexample = counterexample

  def __bool__(self):
    return self.counterexample is None

class IsEquivalentResult:
  def __init__(self, left_counterexample : Optional[str] = None, right_counterexample : Optional[str] = None):
    self.left_counterexample = left_counterexample
    self.right_counterexample = right_counterexample

  def __bool__(self):
    return self.left_counterexample is None and self.right_counterexample is None


class DFA:
    Transition = Dict[Tuple[State, Character], State]
    type_var = TypeVar('type_var')

    def __init__(self, states: Set[State],
                   characters: Set[Character],
                   transition: Transition,
                         init: State,
                        final: Set[State]):

        self.states = states
        self.characters = characters
        self.transition = transition
        self.init = init
        self.final = final
        self.check()

    def check(self):
        assert len(self.states) > 0, "empty automaton"

        for (state, character) in self.transition:
            assert state in self.states, "unknown state " + state.name
            assert character in self.characters, "unknown character " + character.name
            assert self.transition[state, character] in self.states, \
            "unknown state " + self.transition[state, character].name

        assert self.init in self.states, "init not in states"
        for state in self.final:
            assert state in self.states, "unknown state " + state.name

    def total(self) -> DFA:
        if self.is_total():
            return self

        dfa = deepcopy(self)
        total: DFA.Transition = {}

        # check: consider import itertools
        hell_id = 0
        hell = State(str(hell_id))
        while hell in dfa.states:
            hell_id += 1
            hell.name = str(hell_id)
        dfa.states.add(hell)

        for state in dfa.states:
            for character in dfa.characters:
                if (state, character) in dfa.transition:
                    total[state, character] = dfa.transition[state, character]
                else:
                    total[state, character] = hell

        dfa.transition = total
        return dfa

    def is_total(self) -> bool:
        for state in self.states:
            for character in self.characters:
                if (state, character) not in self.transition:
                    return False
        return True

    def accepts(self, word: Union[str, List[Character]]) -> bool:
        evaluate: List[Character] = []
        if isinstance(word, str):
            for letter in word:
                character = Character(letter)
                evaluate.append(character)
        else:
            evaluate = word

        state = self.init
        for character in evaluate:
            if (state, character) not in self.transition:
                return False
            state = self.transition[state, character]
        return state in self.final

    def complement(self) -> DFA:
        dfa = deepcopy(self)
        if not dfa.is_total():
            dfa = dfa.total()

        dfa.final = dfa.states.difference(dfa.final)
        return dfa

    @staticmethod
    def state_composition(state_1: State, state_2: State) -> State:
        return State(f"{state_1.name}_{state_2.name}")

    @staticmethod
    def composition(dfa_1: DFA, dfa_2: DFA, operation: Composition) -> DFA:
        """Does not require input automata to have the same alphabet."""

        # new characters
        characters = dfa_1.characters.union(dfa_2.characters)
        dfa_1.characters = characters
        dfa_2.characters = characters

        if not dfa_1.is_total():
            dfa_1 = dfa_1.total()
        if not dfa_2.is_total():
            dfa_2 = dfa_2.total()

        # new init
        init = DFA.state_composition(dfa_1.init, dfa_2.init)

        # new states and transition
        states = {init}
        transition: Dict[Tuple[State, Character], State] = dict()
        for state_1 in dfa_1.states:
            for state_2 in dfa_2.states:
                state = DFA.state_composition(state_1, state_2)  # TODO avoid conflicts!
                states.add(state)

                # new transition
                for character in characters:
                    if (state_1, character) in dfa_1.transition and \
                            (state_2, character) in dfa_2.transition:
                        dest_state_1 = dfa_1.transition[state_1, character]
                        dest_state_2 = dfa_2.transition[state_2, character]
                        dest_state = DFA.state_composition(dest_state_1, dest_state_2)
                        transition[state, character] = dest_state

        # new final
        final: Set[State] = set()
        for state_1 in dfa_1.final:
            for state_2 in dfa_2.final:
                state = DFA.state_composition(state_1, state_2)

                if operation == Composition.Union:
                    final.add(state)

                elif operation == Composition.Intersection and \
                        state_1 in dfa_1.final and \
                        state_2 in dfa_2.final:
                    final.add(state)

                elif operation == Composition.Subtraction and \
                        state_2 not in dfa_2.final:
                    final.add(state)

        return DFA(states, characters, transition, init, final)

    @staticmethod
    def union(dfa_1: DFA, dfa_2: DFA) -> DFA:
        return DFA.composition(dfa_1, dfa_2, Composition.Union)

    @staticmethod
    def intersection(dfa_1: DFA, dfa_2: DFA) -> DFA:
        return DFA.composition(dfa_1, dfa_2, Composition.Intersection)

    @staticmethod
    def subtraction(dfa_1: DFA, dfa_2: DFA) -> DFA:
        return DFA.composition(dfa_1, dfa_2, Composition.Subtraction)

    def eliminate_unreachable(self) -> DFA:
        dfa = deepcopy(self)
        previous: Set[State] = set()
        actual = {self.init}

        while previous != actual:
            previous = deepcopy(actual)

            for state in previous:
                for character in dfa.characters:
                    if (state, character) in self.transition:
                        actual.add(self.transition[state, character])

        dfa.states = actual
        dfa.final = dfa.final.intersection(actual)
        for (state, character) in self.transition:
            if state not in actual:
                del dfa.transition[state, character]

        return dfa

    def sorted_characters(self) -> List[Character]:
        return sorted(self.characters, key=lambda x: x.name)

    def minimize(self) -> DFA:
        dfa = self.eliminate_unreachable().total()
        characters = self.sorted_characters()  # bc I want to iterate always in same order

        previous: Dict[State, int] = dict()
        actual: Dict[State, int] = dict()
        for state in dfa.states:
            if state in dfa.final:
                actual[state] = 1
            else:
                actual[state] = 0

        classes_prev = 0
        classes_new = 1 if len(self.final) > 0 else 0
        if len(self.states.difference(self.final)) > 0:
            classes_new += 1

        while classes_prev != classes_new:
            previous = deepcopy(actual)
            patterns: Dict[Tuple[int, ...], Set[State]] = dict()
            actual = dict()

            for state in dfa.states:
                pattern = [previous[state]]
                for character in characters:
                    pattern.append(previous[dfa.transition[state, character]])

                patterns.setdefault(tuple(pattern), set()).add(state)

            classes_prev = classes_new
            classes_new = 0
            for states in patterns.values():
                for state in states:
                    actual[state] = classes_new
                classes_new += 1

        new_transition: DFA.Transition = {}
        dfa.states = set()
        dfa.final = set()
        for state in actual:
            new_state = State("new_" + str(actual[state]))
            dfa.states.add(new_state)

            for character in characters:
                if not (new_state, character) in new_transition:
                    old = dfa.transition[state, character]
                    now = State("new_" + str(actual[old]))
                    new_transition[new_state, character] = now

            if state == self.init:
                dfa.init = new_state
            if state in self.final:
                dfa.final.add(new_state)

        dfa.transition = new_transition
        return dfa

    def rename_state(self, rename: State, name: str):
        if name == rename.name:
            return
        new_state = State(name)
        self.states.add(new_state)
        self.states.remove(rename)
        transition = deepcopy(self.transition)

        if rename == self.init:
            self.init = new_state
        if rename in self.final:
            self.final.add(new_state)
            self.final.remove(rename)

        for character in self.characters:
            if (rename, character) in self.transition:
                transition[new_state, character] = self.transition[rename, character]
                del transition[rename, character]

        for state, character in transition:
            if transition[state, character] == rename:
                transition[state, character] = new_state

        self.transition = transition

    def canonize(self) -> DFA:
        dfa = deepcopy(self)
        remain = deepcopy(self.states)
        characters = dfa.sorted_characters()

        i = 0
        queue = deque([dfa.init])
        while len(queue) > 0:
            actual = queue.popleft()
            if actual in remain:
                remain.remove(actual)

            for character in characters:
                if (actual, character) in dfa.transition and dfa.transition[actual, character] in remain:
                    queue.append(dfa.transition[actual, character])
                    remain.remove(dfa.transition[actual, character])

            dfa.rename_state(actual, ascii_uppercase[i])
            i += 1
        return dfa

    def dfa_to_reggrammar(self) -> Any:
        from grammars import RegGrammar

        nonterminals: Set[Nonterminal] = set()
        terminals: Set[Terminal] = set()
        rules: Dict[Nonterminal, Set[Union[Terminal, Tuple[Terminal, Nonterminal]]]] = dict()

        for state in self.states:
            nonterminals.add(Nonterminal(state.name))
        for character in self.characters:
            terminals.add(Terminal(character.name))

        init = Nonterminal("new_init")  # TODO assert that name is unique, as for hell in make_total
        nonterminals.add(init)

        for state, character in self.transition:
            dest_state = self.transition[state, character]
            if state == self.init:
                nonterminal2 = Nonterminal(dest_state.name)
                terminal = Terminal(character.name)

                rules.setdefault(init, set()).add((terminal, nonterminal2))

                if dest_state in self.final:
                    rules[init].add(terminal)
            else:
                nonterminal1 = Nonterminal(state.name)
                nonterminal2 = Nonterminal(dest_state.name)
                terminal = Terminal(character.name)

                rules.setdefault(nonterminal1, set()).add((terminal, nonterminal2))

                if dest_state in self.final:
                    rules[nonterminal1].add(terminal)

        reg = RegGrammar(nonterminals, terminals, rules, init)
        if self.init in self.final:
            pass  # TODO eg.rules[init].add((\eps))

        return reg

    def is_empty(self) -> IsEmptyResult:
        reached: Deque[State] = deque([self.init])
        reachable: Set[State] = {self.init}
        predecessor: Dict[State, Tuple[State, Character]] = {}

        while len(reached) > 0:
            actual = reached.popleft()
            for character in self.characters:
                if (actual, character) in self.transition:
                    dest_state = self.transition[actual, character]
                    if dest_state not in reachable:
                        reached.append(self.transition[actual, character])
                        reachable.add(self.transition[actual, character])
                        predecessor[dest_state] = (actual, character)

        if len(reachable.intersection(self.final)) == 0:
            return IsEmptyResult()

        word = []
        state = reachable.intersection(self.final).pop()
        if state == self.init:
            return IsEmptyResult('ε') # TODO some Eps -> str

        else:
            while state is not self.init:
                word.append(predecessor[state][1].name)
                state = predecessor[state][0]

            counterexample = IsEmptyResult(''.join(reversed(word)))
            return counterexample

    # Ha!
    def is_universal(self):
        return self.complement().is_empty()

    # returns: True or fst_empty: what IS in first language and NOT in second, snd_empty: analogically)
    @staticmethod
    def is_equivalent(fst: DFA, snd: DFA) -> IsEquivalentResult:
        fst_empty = DFA.intersection(fst, snd.complement()).is_empty()
        snd_empty = DFA.intersection(fst.complement(), snd).is_empty()

        if fst_empty and snd_empty:
            result = IsEquivalentResult()
        else:
            result = IsEquivalentResult(fst_empty.counterexample, snd_empty.counterexample)
        return result

    @staticmethod
    def is_part_identical(fst: DFA, snd: DFA) -> bool:
        return fst.states == snd.states and fst.characters == snd.characters and \
               fst.transition == snd.transition and fst.init == snd.init and \
               fst.final == snd.final

    def is_minimal(self) -> bool:
        minimal = self.minimize()
        return self.is_total() and len(self.states) == len(minimal.states) and \
               len(self.transition) == len(minimal.transition)

    def is_canonical(self) -> bool:
        canonic = self.canonize()
        return DFA.is_part_identical(self, canonic)
