from typing import Tuple
import reg
import sys
import signal


def get_task(string: str) -> Tuple[str, str]:
    assert len(string) >= 7, "Teacher solution could not be parsed correctly."
    teacher_type, task = string.split("-", 1)
    if len(task) > 3:
        task, _ = task.split("-", 1)
    assert len(teacher_type) == 3 and len(task) == 3, "Teacher solution could not be parsed correctly."

    return (teacher_type, task)


def dfa_transform(string: str, automaton_type: str) -> reg.DFA:
    try:
        parser = reg.Parser()
        if automaton_type in {"DFA", "TOT", "MIN", "TOC", "MIC"}:
            automaton = parser.str_to_dfa(string)
        elif automaton_type == "NFA":
            automaton = parser.str_to_nfa(string).determinize()
        elif automaton_type == "EFA":
            automaton = parser.str_to_nfa(string).eliminate_epsilon().determinize()
        elif automaton_type == "GRA":
            automaton = parser.str_to_reggrammar(string).reggrammar_to_nfa().determinize()
        elif automaton_type == "REG":
            # test prints
            # regex = parser.str_to_regex(string)
            # efa = regex.regex_to_efa()
            # nfa = efa.eliminate_epsilon()
            # dfa = nfa.determinize()
            # print(parser.regex_to_str(regex))
            # print(parser.nfa_to_str(efa))
            # print(parser.nfa_to_str(nfa))
            # print(parser.dfa_to_str(dfa))

            automaton = parser.str_to_regex(string).regex_to_efa().eliminate_epsilon().determinize()
        return automaton

    except reg.ParsingError as ex:
        print("Chyba při parsování.")
        exit(1)
    except Exception as ex:
        print("Chyba při parsování.")
        exit(1)


def nfa_transform(string: str, automaton_type: str) -> reg.NFA:
    try:
        parser = reg.Parser()
        return parser.str_to_nfa(string)

    except reg.Parser.ParsingError as message:
        print("Parsing error:", message)
        exit(1)


def check_task(dfa: reg.DFA, task: str) -> bool:
    if task not in {"TOT", "MIN", "TOC", "MIC"}:
        return True

    if task == "TOT" and not dfa.is_total():
        print("DFA není totální.")
        return False

    if task == "MIN" and not dfa.is_minimal():
        print("DFA není minimální.")
        return False

    canonical = dfa.is_canonical()
    if task == "CAN" and not canonical:
        print("DFA není kanonický.")
        return False

    if task == "TOC" and (not dfa.is_total() or not canonical):
        if not dfa.is_total():
            print("DFA není totální.")
        if not canonical:
            print("DFA není kanonický.")
        return False

    if task == "MIC" and (not dfa.is_minimal() or not canonical):
        if not dfa.is_minimal():
            print("DFA není minimální.")
        if not canonical:
            print("DFA není kanonický.")
        return False

    return True


def main():
    signal.alarm(60)

    # Get string with student's solution.
    with open(sys.argv[1]) as student_file:
        student_string = student_file.read()

    # Get string with teacher's solution and type of the task.
    for argument in sys.argv:
        if argument[:2] == "-o":
            teacher_string = argument[2:]

    try:
        task_prefix, teacher_string = teacher_string.split(":", 1)
        teacher_type, task = get_task(task_prefix)

        student_solution = dfa_transform(student_string, task)
        teacher_solution = dfa_transform(teacher_string, teacher_type)

        #
        if teacher_solution.is_empty() and student_solution.is_empty():
            print("Správné řešení.")
            exit(0)

        elif student_solution.characters != teacher_solution.characters:
            # TODO ask also print both alphabets?
            # reg.Parser.names_to_str(student_solution.characters)

            if task == "REG":
                print("Abecedy zadaného jazyka a studentova řešení se liší "
                    "(abeceda je odvozena ze všech písmen objevujících se v regulárním výrazu).")
            elif task == "GRA":
                print("Abecedy zadaného jazyka a studentova řešení se liší "
                    "(abeceda je odvozena ze všech terminálů objevujících se v pravidlech gramatiky).")
            else:
                print("Abecedy zadaného jazyka a studentova řešení se liší "
                    "(abeceda je odvozena ze všech písmen objevujících se v přechodové funkci).")
            print("Nesprávné řešení.")
            exit(1)

        result = reg.DFA.is_equivalent(student_solution, teacher_solution)
        if check_task(student_solution, task) and result:
            print("Správné řešení.")
            exit(0)

        if not result:
            student_word, teacher_word = result.left_counterexample, result.right_counterexample

            if student_word is None:
                print("Studentovo řešení je podmnožinou zadaného jazyka.")
            else:
                print("Příklad slova, které je ve studentově řešení a není v zadaném jazyce: ", student_word)

            if teacher_word is None:
                print("Studentovo řešení je nadmnožinou zadaného jazyka.")
            else:
                print("Příklad slova, které chybí ve studentově řešení a je v zadaném jazyce: ", teacher_word)

        print("Nesprávné řešení.")
        exit(1)

    except Exception as ex:
    #except AttributeError as message:
        print("Error inside of fja_checker:", ex.args)
        exit(1)

main()